/* $Id: fflib.c,v 3.1 1996/11/02 23:31:29 kas Exp $ */

/* Pomocne funkce pro praci se silovym polem */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995, 1996  Jan Kasprzak <kas@muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#define DEBUG 1

#ifdef DEBUG
#include <stdio.h>
#endif

#include <stdlib.h>

#include "ff.h"


/* Funkce naalokuje dynamicke casti force_field podle dimenzi,
 * uvedenych ve force_field (krome ffp->title, ktere ma dynamickou
 * delku. Prislusne celociselne prvky struktury force_field
 * musi byt v dobe volani funkce definovany (ffp->natoms,nbondtypes, ...).
 */
void malloc_ff(ffp)
force_field *ffp;
{
	int i;

	ffp->anames = malloc(sizeof(char *)*ffp->natoms);
	ffp->atypenames = malloc(sizeof(char *)*ffp->natoms);
	for (i=0; i<ffp->natoms; i++)
	{
		ffp->anames[i] = malloc(ATOM_NAME_LENGTH*sizeof(char));
		ffp->atypenames[i] = malloc(ATOM_NAME_LENGTH*sizeof(char));
	}

	ffp->charge = malloc(sizeof(REAL)*ffp->natoms);
	ffp->r12 = malloc(sizeof(REAL)*(ffp->natypes+1)*ffp->natypes/2);
	ffp->r6 = malloc(sizeof(REAL)*(ffp->natypes+1)*ffp->natypes/2);

	if (ffp->nhydronb > 0) {
		ffp->h12 = malloc(sizeof(REAL)*(ffp->nhydronb));
		ffp->h10 = malloc(sizeof(REAL)*(ffp->nhydronb));
	} else
		ffp->h12 = ffp->h10 = (REAL *)0;

	ffp->rk = malloc(sizeof(REAL)*ffp->nbondtypes);
	ffp->req = malloc(sizeof(REAL)*ffp->nbondtypes);

	ffp->tk = malloc(sizeof(REAL)*ffp->nangletypes);
	ffp->teq = malloc(sizeof(REAL)*ffp->nangletypes);

	ffp->pk = malloc(sizeof(REAL)*ffp->ndihtypes);
	ffp->pn = malloc(sizeof(int)*ffp->ndihtypes);
	ffp->phase = malloc(sizeof(int)*ffp->ndihtypes);

	ffp->atypes = malloc(sizeof(int)*ffp->natoms);
	ffp->nbnum = malloc(sizeof(int)*ffp->natypes*ffp->natypes);

	ffp->bondat1 = malloc(sizeof(int)*ffp->nbonds);
	ffp->bondat2 = malloc(sizeof(int)*ffp->nbonds);
	ffp->bondnum = malloc(sizeof(int)*ffp->nbonds);

	ffp->angleat1 = malloc(sizeof(int)*ffp->nangles);
	ffp->angleat2 = malloc(sizeof(int)*ffp->nangles);
	ffp->angleat3 = malloc(sizeof(int)*ffp->nangles);
	ffp->anglenum = malloc(sizeof(int)*ffp->nangles);

	ffp->dihat1 = malloc(sizeof(int)*ffp->ndihs);
	ffp->dihat2 = malloc(sizeof(int)*ffp->ndihs);
	ffp->dihat3 = malloc(sizeof(int)*ffp->ndihs);
	ffp->dihat4 = malloc(sizeof(int)*ffp->ndihs);
	ffp->dihnum = malloc(sizeof(int)*ffp->ndihs);
	ffp->improper = malloc(sizeof(int)*ffp->ndihs);
	ffp->ignored = malloc(sizeof(int)*ffp->ndihs);

#ifndef NONTRANSFORMED_ONLY
	ffp->tdihnum = malloc(sizeof(int)*ffp->ndihs);
	ffp->tanglenum = malloc(sizeof(int)*ffp->nangles);

	ffp->lambda = -1.0;
#endif
	/* Default values */
	ffp->wdw_14 = 2.0;
	ffp->eel_14 = 1.2;
	ffp->diel_lin = 0.0;
	ffp->diel_const = 1.0;
	ffp->cutoff_sq = -1.0;
}

/* Funkce uvolni vsechny dynamicke prvky ze struktury force_field
 * i strukturu samotnou.
 */
void free_ff(ffp)
force_field *ffp;
{
	int i;

	free(ffp->title);	/* This is not malloc'd in malloc_ff() */

	for (i=0; i<ffp->natoms; i++)
	{
		free(ffp->anames[i]);
		free(ffp->atypenames[i]);
	}

	free(ffp->anames);
	free(ffp->atypenames);

	free(ffp->charge);
	free(ffp->r12);
	free(ffp->r6);

	if (ffp->h12)
		free(ffp->h12);
	if (ffp->h10)
		free(ffp->h10);

	free(ffp->rk);
	free(ffp->req);

	free(ffp->tk);
	free(ffp->teq);

	free(ffp->pk);
	free(ffp->pn);
	free(ffp->phase);

	free(ffp->atypes);
	free(ffp->nbnum);
	
	free(ffp->bondat1);
	free(ffp->bondat2);
	free(ffp->bondnum);

	free(ffp->angleat1);
	free(ffp->angleat2);
	free(ffp->angleat3);
	free(ffp->anglenum);

	free(ffp->dihat1);
	free(ffp->dihat2);
	free(ffp->dihat3);
	free(ffp->dihat4);
	free(ffp->dihnum);
	free(ffp->improper);
	free(ffp->ignored);
#ifndef NONTRANSFORMED_ONLY
	free(ffp->tdihnum);
	free(ffp->tanglenum);

	if (ffp->lambda < 0.0)
		free_trans_part(ffp);
#endif
	free(ffp);
}

/* Funkce find_bond() vraci typ vazby mezi zadanymi dvema atomy,
 * pokud tato vazba existuje. Pokud neexistuje, vraci -1.
 */
int find_bond(ffp, a1, a2)
force_field *ffp;
int a1, a2;
{
        int i;
	for (i=0; i<ffp->nbonds; i++)
		if ((ffp->bondat1[i] == a1 && ffp->bondat2[i] == a2)
			|| (ffp->bondat2[i] == a1 && ffp->bondat1[i] == a2))
			return ffp->bondnum[i];
#ifdef DEBUG
        fprintf(stderr, "WARNING: Bond between %d and %d not found!\n", a1, a2);
#endif
	return -1;
}

/* Funkce find_angle() funguje podobne jako find_bond() -- vraci typ
 * uhlu mezi tremi zadanymi atomy.
 */
int find_angle(ffp, a1, a2, a3)
force_field *ffp;
int a1, a2, a3;
{
        int i;
        for (i=0; i<ffp->nangles; i++)
                if ((ffp->angleat2[i] == a2) &&
                        ((ffp->angleat1[i] == a1 && ffp->angleat3[i] == a3)
                        || (ffp->angleat3[i] == a1 && ffp->angleat1[i] == a3)))
			return ffp->anglenum[i];

#ifdef DEBUG
	fprintf(stderr, "WARNING: Angle %d-%d-%d not found!\n", a1, a2, a3);
#endif
	return -1;
}

void setup_ff(ffp, wdw14, eel14, diel_lin, diel_const, cutoff)
force_field *ffp;
REAL wdw14, eel14, diel_lin, diel_const, cutoff;
{
	if (wdw14 >= 0.0)
		ffp->wdw_14 = wdw14;
	if (eel14 >= 0.0)
		ffp->eel_14 = eel14;
	if (diel_lin >= 0.0)
		ffp->diel_lin = diel_lin;
	if (diel_const >= 0.0)
		ffp->diel_const = diel_const;
	if (cutoff >= 0.0)
		ffp->cutoff_sq = cutoff*cutoff;
}

