/* $Id: printangle.c,v 3.0.0.1 1996/11/02 18:12:38 kas Exp $ */

/* Program pro vypis parametru uhlu u molekuly.
 * Vystup tohoto programu ma slouzit jako vstup pro vypocet tabulek
 * ,,vyhlazene`` silove funkce pro uhly.
 */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995, 1996  Jan Kasprzak <kas@muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <unistd.h>
#include <getopt.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "ff.h"

#define DEFAULT_DIMENSION	24.0	/* Tabulky po 15 stupnich */
#define ATOM_NAME	"%.2s"
#define RNUM		"% 12.8f"

static char *program_name;

static void usage __P(());
static void die __P((char *msg));
void print_angle_parms __P((FILE *f, force_field *ffp, int adim));
	
int main(argc, argv)
int argc;
char **argv;
{
	int adim = 0;
	force_field *ffp;

	program_name = *argv;

	while(1)
	{
		int c, opt_ind;
		static struct option long_opts[] = {
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'v' },
			{ "dimension", 1, 0, 'd' }, 
		};

		if ((c=getopt_long(argc, argv, "hvd:--", long_opts,
			&opt_ind)) == -1)
			break;
		switch(c)
		{
		case 'h':
		case '?':
			usage();
		case 'v':
			printf ("Version: $Id: printangle.c,v 3.0.0.1 1996/11/02 18:12:38 kas Exp $\n");
			return 0;
		case 'd':
			if (!adim)
				adim = atoi(optarg);
			else
				die("angle-dimension set more than once!");
		}
	}

	if (optind+1 > argc)
		usage();
	
	if ((ffp = prm2ff(argv[optind])) == NULL)
	{
		perror ("Opening force field file");
		exit (3);
	}

	if (!adim)
		adim = DEFAULT_DIMENSION;

	print_angle_parms(stdout, ffp, adim);

	return 0;
}

/* Funkce vypise parametry uhlu a souradnice prislusnych bodu.
 * Body jsou zvoleny v jiste "normalizovane" poloze:
 * Bod 2 je v pocatku souradnic, bod 1 na ose x v kladnem smeru,
 * bod 3 v rovine xy s kladnym y.
 */
void print_angle_parms(out, ffp, adim)
FILE *out;
force_field *ffp;
int adim;
{
	REAL theta;
	REAL theta0;	/* Equillibrium values */
	REAL l1, l2;
	int a, i;
	for (a=0; a<ffp->nangles; a++)
	{
		l1 = ffp->req[find_bond(ffp, ffp->angleat1[a],
				ffp->angleat2[a])];
		l2 = ffp->req[find_bond(ffp, ffp->angleat2[a],
				ffp->angleat3[a])];
		theta0 = ffp->teq[a];

		fprintf(out, ATOM_NAME "-" ATOM_NAME "-" ATOM_NAME " " RNUM " %d " RNUM " " RNUM "\n",
			ffp->atypenames[ffp->angleat1[a]],
			ffp->atypenames[ffp->angleat2[a]],
			ffp->atypenames[ffp->angleat3[a]],
			ffp->teq[ffp->anglenum[a]],
			adim, l1, l2);
#ifdef START_ON_EQUILLIBRIUM_ANGLE
		for (i=0,theta=theta0; i<adim; i++,phi+=M_PI/(adim-1))
#else
		for (i=0,theta=0.0; i<adim; i++,theta+=M_PI/(adim-1))
#endif
		{
			fprintf(out, RNUM " ", theta);
			/* Bod 1 */
			fprintf(out, RNUM " " RNUM " " RNUM " ",
				l1, 0.0, 0.0);
			/* Bod 2 */
			fprintf(out, RNUM " " RNUM " " RNUM " ",
				0.0, 0.0, 0.0);
			/* Bod 3 */
			fprintf(out, RNUM " " RNUM " " RNUM "\n",
				l2*cos(theta), l2*sin(theta), 0.0);
		}
	}
}

static void usage()
{
	printf("Usage: %s [option...] top_file\n"
		"Where option is one or more from:\n"
		"\t-d <number>\tSets the number of points generated\n"
		"\t-v or --version\tPrints the RCS version of program\n"
		"\t-h or --help\tPrints help\n"
		"\t--dimension\tSame as -d\n"
		, program_name);
	exit (1);
}

static void die(s)
char *s;
{
	fprintf(stderr, "%s\nExiting.\n", s);
	exit(2);
}

