/* $Id: pval.c,v 3.0.0.1 1996/11/02 18:12:38 kas Exp $ */

/* Rutiny pro vypocet funkcni hodnoty potencialove funkce */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995, 1996  Jan Kasprzak <kas@muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "ff.h"
#include "pval.h"

#define PRINT_DEBUG 1

potential_value *alloc_pv(f)
force_field *f;
{
	potential_value *pvp;
	int i,j,s, *tmp, *nb, mx;
	int maxangles = 0, maxbonds = 0, maxdihs = 0;

	pvp = malloc(sizeof(potential_value));

	pvp->ffp = f;

	pvp->nbval = malloc(sizeof(REAL)*f->natoms*f->natoms);
	pvp->bondval = malloc(sizeof(REAL)*f->nbonds);
	pvp->angleval = malloc(sizeof(REAL)*f->nangles);
	pvp->dihval = malloc(sizeof(REAL)*f->ndihs);

	pvp->x = malloc(sizeof(REAL)*f->natoms);
	pvp->y = malloc(sizeof(REAL)*f->natoms);
	pvp->z = malloc(sizeof(REAL)*f->natoms);

	pvp->nbonds = malloc(sizeof(int *)*f->natoms);
	pvp->bonds = malloc(sizeof(int *)*f->natoms);
	pvp->angles = malloc(sizeof(int *)*f->natoms);
	pvp->dihs = malloc(sizeof(int *)*f->natoms);

	/* Done allocating basic arrays, now we have to compute
	 * bonds, angles and dihs arrays for each atom.
	 */
#define max(x,y)	((x)<(y) ? (y) : (x))
	mx = max(f->natoms, f->nbonds);
	mx = max(f->nangles, mx);
	mx = max(f->ndihs, mx);

	tmp = malloc(sizeof(int)*mx);
	nb = malloc(sizeof(int)*(f->natoms));

	for (i=0; i<f->natoms; i++)
	{
		for (j=0; j<f->natoms; j++)
			nb[j] = 1;
		nb[i] = 0;

		for (j=s=0; j<f->nbonds; j++)
			if (f->bondat1[j] == i || f->bondat2[j] == i)
			{
				tmp[s++] = j;
				nb[f->bondat1[j]] = 0;
				nb[f->bondat2[j]] = 0;
			}
		pvp->bonds[i] = malloc(sizeof(int)*(s+1));
		pvp->bonds[i][0] = s;
		if (s)
			memcpy(pvp->bonds[i]+1,tmp,s*sizeof(int));

		if (s > maxbonds)
			maxbonds = s;

		for (j=s=0; j<f->nangles; j++)
			if (f->angleat1[j] == i || f->angleat2[j] == i
				|| f->angleat3[j] == i)
			{
				tmp[s++] = j;
				nb[f->angleat1[j]] = 0;
				nb[f->angleat2[j]] = 0;
				nb[f->angleat3[j]] = 0;
			}
		pvp->angles[i] = malloc(sizeof(int)*(s+1));
		pvp->angles[i][0] = s;
		if (s)
			memcpy(pvp->angles[i]+1,tmp,s*sizeof(int));

		if (s > maxangles)
			maxangles = s;

		for (j=s=0; j<f->ndihs; j++)
			if (f->dihat1[j] == i || f->dihat2[j] == i ||
				f->dihat3[j] == i || f->dihat4[j] == i)
			{
				tmp[s++] = j;
				nb[f->dihat1[j]] = 0;
				nb[f->dihat2[j]] = 0;
				nb[f->dihat3[j]] = 0;
				nb[f->dihat4[j]] = 0;
			}
		pvp->dihs[i] = malloc(sizeof(int)*(s+1));
		pvp->dihs[i][0] = s;
		if (s)
			memcpy(pvp->dihs[i]+1,tmp,s*sizeof(int));

		if (s > maxdihs)
			maxdihs = s;

		/* Now find all nonbonded interactions for atom i */
		for(j=0, s=0; j<f->natoms; j++)
			if (nb[j])
				s++;
		pvp->nbonds[i] = malloc(sizeof(int)*(s+1));
		pvp->nbonds[i][0] = s;
		for(j=0, s=0; j<f->natoms; j++)
			if (nb[j])
				pvp->nbonds[i][++s] = j;
	}
	free(tmp);
	free(nb);
	pvp->recomputed_atom = -1;

#ifdef DEBUG
	printf("maxbonds = %d, maxangles = %d, maxdihs = %d\n",
		maxbonds, maxangles, maxdihs);
#endif
	pvp->onbval = malloc(sizeof(REAL)*f->natoms);
	pvp->obondval = malloc(sizeof(REAL)*maxbonds);
	pvp->oangleval = malloc(sizeof(REAL)*maxangles);
	pvp->odihval = malloc(sizeof(REAL)*maxdihs);

	return pvp;
}

void free_pv(pvp)
potential_value *pvp;
{
	int i;

	free(pvp->bondval);
	free(pvp->angleval);
	free(pvp->dihval);

	free(pvp->x);
	free(pvp->y);
	free(pvp->z);

	free(pvp->obondval);
	free(pvp->oangleval);
	free(pvp->odihval);

	for (i=0; i<pvp->ffp->natoms; i++)
	{
		free(pvp->bonds[i]);
		free(pvp->angles[i]);
		free(pvp->dihs[i]);
	}

	free(pvp->bonds);
	free(pvp->angles);
	free(pvp->dihs);

	free_ff(pvp->ffp);
}

int read_coordinates(pvp, filename)
potential_value *pvp;
char *filename;
{
	FILE *f;

	int i;

	if ((f = fopen(filename, "r")) == NULL)
		return 0;
	
	for (i=0; i<pvp->ffp->natoms; i++)
		if (fscanf(f, "%lf %lf %lf \n",
			pvp->x+i, pvp->y+i, pvp->z+i) != 3)
		{
			fclose(f);
			return 0;
		}

	fclose(f);

	return 1;
}

int read_crd_file(pvp, filename)
potential_value *pvp;
char *filename;
{
	FILE *f;
	char buf[2560];
	int i;

	if ((f = fopen(filename, "r")) == NULL)
		return 0;
	
	/* First, we must read the name of the molecule */
	fgets (buf, 255, f);

	/* Then we must read and check # of atoms */
	fscanf(f, "%d", &i);

	if (i != pvp->ffp->natoms)
	{
		fprintf(stderr, "read_crd_file(): Wrong # of atoms\n");
		return 0;
	}

	for (i=0; i<pvp->ffp->natoms; i++)
		if (fscanf(f, "%lf %lf %lf",
			pvp->x+i, pvp->y+i, pvp->z+i) != 3)
		{
			fclose(f);
			return 0;
		}

	fclose(f);

	return 1;
}


int write_crd_file(pvp, filename, x, y, z)
potential_value *pvp;
char *filename;
REAL *x, *y, *z;
{
	FILE *f;
	int i, odd;

	if ((f = fopen(filename, "w")) == NULL)
		return 0;
	
	fputs(pvp->ffp->title, f);
	fputc('\n', f);

	fprintf(f, "%4d\n", pvp->ffp->natoms);

	for (i=0, odd=0; i<pvp->ffp->natoms; i++, odd=!odd)
	{
		fprintf(f, " % 11.7f % 11.7f % 11.7f",
			 x[i], y[i], z[i]);
		if (odd)
			fputc('\n', f);
	}
	fclose(f);
	return 1;
}

int write_coordinates(pvp, filename)
potential_value *pvp;
char *filename;
{
	int i;
	FILE *f;

	if ((f = fopen(filename, "w")) == NULL)
		return 0;

	for (i=0; i < pvp->ffp->natoms; i++)
		if (fprintf(f, "%f %f %f\n", pvp->x[i], pvp->y[i],
			pvp->z[i]) == 0)
		{
			fclose(f);
			return 0;
		}
	
	if (fclose(f))
		return 0;
	return -1;
}

REAL compute_potential(pvp)
potential_value *pvp;
{
	int i,j;
#ifdef PRINT_DEBUG
	REAL r = 0.0;
#endif
	pvp->value = 0.0;

	for (i=0; i<pvp->ffp->nbonds; i++)
		pvp->value += pvp->bondval[i] = compute_bond(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
#ifdef PRINT_DEBUG
	printf ("Bonds total: %3.3f\n", pvp->value);
	r = pvp->value;
#endif
	for (i=0; i<pvp->ffp->nangles; i++)
		pvp->value += pvp->angleval[i] = compute_angle(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
#ifdef PRINT_DEBUG
	printf ("Angles total: %3.3f\n", pvp->value-r);
	r = pvp->value;
#endif
	for (i=0; i<pvp->ffp->ndihs; i++)
		pvp->value += pvp->dihval[i] = compute_dih(pvp->ffp,
			pvp->x, pvp->y, pvp->z, i);
#ifdef PRINT_DEBUG
	printf ("Dihedrals total: %3.3f\n", pvp->value-r);
	r = pvp->value;
#endif
	for (i=0; i<pvp->ffp->natoms; i++)
		for (j=0; j<pvp->nbonds[i][0]; j++)
		{
			REAL foo1, foo2;

			if (pvp->nbonds[i][1+j] <= i)
				continue;
			pvp->value +=
			pvp->nbval[pvp->ffp->natoms*i+pvp->nbonds[i][1+j]] =
				compute_nb(pvp->ffp, pvp->x, pvp->y, pvp->z,
					i, pvp->nbonds[i][1+j], &foo1, &foo2,
					0);
		}
#ifdef PRINT_DEBUG
	printf ("Nonbonded interactions total: %3.3f\n", pvp->value-r);
#endif
	pvp->recomputed_atom = -1;

	return pvp->value;
}

REAL recompute_potential(pvp, atom, nx, ny, nz)
potential_value *pvp;
int atom;
REAL nx, ny, nz;
{
	int i;

	pvp->ox = pvp->x[atom];
	pvp->oy = pvp->y[atom];
	pvp->oz = pvp->z[atom];

	pvp->x[atom] = nx;
	pvp->y[atom] = ny;
	pvp->z[atom] = nz;

	pvp->recomputed_atom = atom;
	pvp->ovalue = pvp->value;

	/* Recompute all bonds, angles, dihedrals for atom */
#ifdef PRINT_DEBUG
	printf ("Recomputing bonds:\n");
#endif
	for (i=pvp->bonds[atom][0]; i; i--)
	{
		int bond;

		bond = pvp->bonds[atom][i];
		pvp->obondval[i-1] = pvp->bondval[bond];
		pvp->value -= pvp->bondval[bond];

		pvp->bondval[bond] = compute_bond(pvp->ffp,
			pvp->x, pvp->y, pvp->z, bond);
		pvp->value += pvp->bondval[bond];
	}
#ifdef PRINT_DEBUG
	printf ("Recomputing angles:\n");
#endif
	for (i=pvp->angles[atom][0]; i; i--)
	{
		int angle;

		angle = pvp->angles[atom][i];
		pvp->oangleval[i-1] = pvp->angleval[angle];
		pvp->value -= pvp->angleval[angle];

		pvp->angleval[angle] = compute_angle(pvp->ffp,
			pvp->x, pvp->y, pvp->z, angle);
		pvp->value += pvp->angleval[angle];
	}
#ifdef PRINT_DEBUG
	printf ("Recomputing dihedrals:\n");
#endif
	for (i=pvp->dihs[atom][0]; i; i--)
	{
		int dih;

		dih = pvp->dihs[atom][i];
		pvp->odihval[i-1] = pvp->dihval[dih];
		pvp->value -= pvp->dihval[dih];

		pvp->dihval[dih] = compute_dih(pvp->ffp,
			pvp->x, pvp->y, pvp->z, dih);
		pvp->value += pvp->dihval[dih];
	}
#ifdef PRINT_DEBUG
	printf ("Recomputing nonbonded interactions:\n");
#endif
	for (i=pvp->nbonds[atom][0]; i; i--)
	{
		int a1,a2, idx;
		REAL foo1, foo2;

		if ((a2 = pvp->nbonds[atom][i]) <= atom) {
			a1 = a2;
			a2 = atom;
		} else
			a1 = atom;
		
		idx = a1*pvp->ffp->natoms + a2;

		pvp->onbval[i-1] = pvp->nbval[idx];
#ifdef PRINT_NB_DEBUG
		printf("onbval=%f\n", pvp->nbval[idx]);
#endif
		pvp->value -= pvp->nbval[idx];

		pvp->nbval[idx] = compute_nb(pvp->ffp, pvp->x, pvp->y,
			pvp->z, a1, a2, &foo1, &foo2, 0);
		pvp->value += pvp->nbval[idx];
#ifdef PRINT_NB_DEBUG
		printf("nbval=%f\n", pvp->nbval[idx]);
		printf("value = %f\n", pvp->value);
#endif
	}

	return pvp->value;
}

REAL undo_new_potential(pvp)
potential_value *pvp;
{
	int i;

	/* This is to discover calling of undo_new_potential() w/o previous
	 * call of recompute_potential().
	 */
	if (pvp->recomputed_atom == -1)
	{
		fprintf(stderr, "Oops! Trying to undo something that did not happen!\n");
		exit(1);
	}

	pvp->value = pvp->ovalue;

	pvp->x[pvp->recomputed_atom] = pvp->ox;
	pvp->y[pvp->recomputed_atom] = pvp->oy;
	pvp->z[pvp->recomputed_atom] = pvp->oz;

	for (i=pvp->bonds[pvp->recomputed_atom][0]; i; i--)
		pvp->bondval[pvp->bonds[pvp->recomputed_atom][i]] =
			pvp->obondval[i-1];

	for (i=pvp->angles[pvp->recomputed_atom][0]; i; i--)
		pvp->angleval[pvp->angles[pvp->recomputed_atom][i]] =
			pvp->oangleval[i-1];

	for (i=pvp->dihs[pvp->recomputed_atom][0]; i; i--)
		pvp->dihval[pvp->dihs[pvp->recomputed_atom][i]] =
			pvp->odihval[i-1];

	for (i=pvp->nbonds[pvp->recomputed_atom][0]; i; i--)
		pvp->nbval [pvp->nbonds[pvp->recomputed_atom][i]
			* pvp->ffp->natoms + pvp->recomputed_atom]
		= pvp->nbval[pvp->recomputed_atom * pvp->ffp->natoms
			+ pvp->nbonds[pvp->recomputed_atom][i]]
		= pvp->onbval[i-1];

	pvp->recomputed_atom = -1;

	return pvp->value;
}

#ifdef DEBUG_COMPUTE
/* The main() function for debugging purposes. */

#include <getopt.h>

static void usage();
static char *program_name;

int main(argc, argv)
int argc;
char **argv;
{
	force_field *ffp;
	potential_value *pvp;
	REAL wdw14 = -1.0, eel14 = -1.0, diel_lin = -1.0, diel_const = -1.0;
	REAL cutoff = -1.0;
#ifndef NONTRANSFORMED_ONLY
	char *pair_template = NULL, *dih_template = NULL, *angle_template = NULL;
	REAL lambda = 0.8;
#endif
	program_name = *argv;

        while(1)
        {
                int c, opt_ind;
                static struct option long_opts[] = {
                        { "help", 0, 0, 'h' },
                        { "version", 0, 0, 'v' },
                        { "wdw14", 1, 0, 'w' },
                        { "eel14", 1, 0, 'e' },
                        { "diel-lin", 1, 0, 'l' },
                        { "diel-const", 1, 0, 'c' },
			{ "cutoff", 1, 0, 'C' },
#ifndef NONTRANSFORMED_ONLY
			{ "pair-template", 1, 0, 'p'},
			{ "dih-template", 1, 0, 'd'},
			{ "angle-template", 1, 0, 'a'},
			{ "lambda", 1, 0, 'L'},
#endif
                };

                if ((c=getopt_long(argc, argv, "hvw:e:l:c:C:--", long_opts,
                        &opt_ind)) == -1)
                        break;
                switch(c)
                {
                case 'h':
                case '?':
                        usage();
                case 'v':
                        printf("Version: $Id: pval.c,v 3.0.0.1 1996/11/02 18:12:38 kas Exp $\n");
                        return 0;
                case 'w':
                        wdw14 = atof(optarg);
                        break;
                case 'e':
                        eel14 = atof(optarg);
                        break;
                case 'l':
                        diel_lin = atof(optarg);
                        break;
                case 'c':
                        diel_const = atof(optarg);
                        break;
		case 'C':
			cutoff = atof(optarg);
			break;
#ifndef NONTRANSFORMED_ONLY
		case 'p':
			pair_template = optarg;
			break;
		case 'a':
			angle_template = optarg;
			break;
		case 'd':
			dih_template = optarg;
			break;
		case 'L':
			lambda = atof(optarg);
			break;
#endif
                default:
                	usage();
                	break;
                }
	}
	if (optind+2 > argc)
		usage();

	ffp = prm2ff(argv[optind]);

        setup_ff(ffp, wdw14, eel14, diel_lin, diel_const, cutoff);
#ifndef NONTRANSFORMED_ONLY
	read_trans_ff(ffp, lambda, angle_template, dih_template, pair_template);
#endif
	pvp = alloc_pv(ffp);
#if 0
	pvp->x[0] = pvp->y[0] = pvp->z[0] = pvp->x[1] = pvp->y[1] = 0.0;
	for (pvp->z[1]=0.05; pvp->z[1] < 3.0; pvp->z[1]+=0.025)
		compute_bond(pvp->ffp, pvp->x, pvp->y, pvp->z, 0);
#else
	read_crd_file(pvp, argv[optind+1]);

	compute_potential(pvp);

	printf ("Total value of potential function = %3.3f\n",
		pvp->value);
#endif
	return 0;
}
static void usage()
{
        fprintf(stderr, "Usage: %s [options] file.top file.crd\n"
                "Where [options] is zero or more from\n"
                "--help\tprints out this message\n"
                "--version\tprints out the version of the program\n"
                "--wdw14 number\tsets the scaling parameter for WdW 1-4 interactions (default is 2.0)\n"
                "--eel14 number\tsets the scaling parameter for E-static 1-4 interactions (1.2)\n"
#ifndef NONTRANSFORMED_ONLY
		"--lambda number\tlambda value for transfomed force field\n"
		"--pair-template string\ttemplate for pair-interactions file\n"
		"--angle-template string\ttemplate for angle-interactions file\n"
		"--dih-template string\ttemplate for torison-interactions file\n"
#endif
                "--diel-lin number\tsets the linear part of dielectric funtion (1.0)\n"
                "--diel-const number\tsets the constant part of dielectric function (0.0)\n"
		"--cutoff number\tsets the cutoff parameter for the 1991 force field (+infinity)\n",
                program_name);
        exit(1);
}

#endif /* DEBUG_COMPUTE */
