/* $Id: translib.c,v 3.2 1996/11/03 12:37:50 kas Exp $ */

/* Rutiny pro praci s tabulkami, popisujicimi transformovane silove pole */

/*
 *  Programs used in my diploma thesis.
 *  Copyright (C) 1995, 1996  Jan Kasprzak <kas@muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define USE_TRANSFORMED	1
#include "ff.h"

#define BUF_SIZE 10240
#define STEP_EPSILON	1e-4
#define LAMBDA_EPSILON	1e-3
#define DEFAULT_PAIRFILE_TEMPLATE	"tabulky/pair_%5.2e"
#define MAX_DIHS	4096
#define DEFAULT_DIHFILE_TEMPLATE	"tabulky/alb_13_%5.2e.gtor"
#define MAX_ANGLES	4096
#define DEFAULT_ANGLEFILE_TEMPLATE	"tabulky/alb_13_%5.2e.gang"

static pair_int *read_pairfile __P((char *tmpl, REAL lambda));
static dih_int *read_dihfile __P((char *tmpl, REAL lambda));
static angle_int *read_anglefile __P((char *tmpl, REAL lambda));

static void free_pip __P((pair_int *pip));
static void free_dip __P((dih_int *dip));
static void free_aip __P((angle_int *aip));

static int read_dip __P((dih_int *dip, FILE *f, REAL lambda));
static int read_aip __P((angle_int *aip, FILE *f, REAL lambda));

static int compute_tdihnum __P((force_field *ffp));
static int compute_tanglenum __P((force_field *ffp));

int read_trans_ff(ffp, lambda, atmpl, ttmpl, ptmpl)
force_field *ffp;
REAL lambda;
char *atmpl, *ttmpl, *ptmpl;
{
	if (lambda < 0.0)
		free_trans_part(ffp);

	ffp->lambda = lambda;
	if ((ffp->pip = read_pairfile(ptmpl, lambda)) == NULL)
		return 0;
	if ((ffp->aip = read_anglefile(atmpl, lambda)) == NULL)
		return 0;
	if ((ffp->dip = read_dihfile(ttmpl, lambda)) == NULL)
		return 0;
	if (!compute_tdihnum(ffp))
		return 0;
	if (!compute_tanglenum(ffp))
		return 0;
	return 1;
}

/* Funkce uvolni slozky force_field, ktere se vazou k transformovanemu poli */
void free_trans_part(ffp)
force_field *ffp;
{
	free_dip(ffp->dip);
	free_aip(ffp->aip);
	free_pip(ffp->pip);
}


/* --------------------------------------------------------------------*/
/* Funkce pro praci s tabulkou parovych interakci */

#define MAX_PAIR_DIMENSION	1024
static pair_int *read_pairfile(template, lambda)
char *template;
REAL lambda;
{
	char buf[BUF_SIZE];
	FILE *f;
	pair_int *pip;
	REAL foo;
	int i, j;

	if (!template)
		template = DEFAULT_PAIRFILE_TEMPLATE;

	sprintf(buf, template, lambda);
	if ((f = fopen(buf, "r")) == NULL) {
		fprintf(stderr, "Cannot open %s: %s\n", buf, strerror(errno));
		return NULL;
	}

	pip = malloc(sizeof(pair_int));

	/* Nacteni hlavicky */
	if (fscanf(f, " n=%*d, lambda=%lf, step=%lf, acc=%*f, eps=%*f, maxpts=%*f",
/*	if (fscanf(f, "lambda=%lf, step=%lf, acc=%*f, eps=%*f, maxpts=%*f", */
		&(pip->lambda), &(pip->step)) != 2
		|| fabs(pip->lambda - lambda) > LAMBDA_EPSILON) {
		fprintf(stderr, "read_pair_int(): Incorrect input\n");
		free(pip);
		return NULL;
	}

	/* Preskocime indexovy radek tabulky */
	fgets(buf, BUF_SIZE, f);
	fgets(buf, BUF_SIZE, f);

	/* Udelame prvni nulovy radek */
	pip->dim=MAX_PAIR_DIMENSION;
	pip->val_m12 = malloc(sizeof(REAL)*pip->dim);
	pip->val_m10 = malloc(sizeof(REAL)*pip->dim);
	pip->val_m6  = malloc(sizeof(REAL)*pip->dim);
	pip->val_m2  = malloc(sizeof(REAL)*pip->dim);
	pip->val_m1  = malloc(sizeof(REAL)*pip->dim);
	pip->val_1   = malloc(sizeof(REAL)*pip->dim);
	pip->val_2   = malloc(sizeof(REAL)*pip->dim);

	for (i=1; i<pip->dim; i++) {
		if (i >= MAX_PAIR_DIMENSION) {
			printf ("MAX_PAIR_DIMENSION (%d) limit reached. Recompile me.\n",
				MAX_PAIR_DIMENSION);
			return NULL;
		}
		j = fscanf(f, " %lf %lf %lf %lf %lf %lf %lf %lf",
			&foo,
			pip->val_m12 + i,
			pip->val_m10 + i,
			pip->val_m6 + i,
			pip->val_m2 + i,
			pip->val_m1 + i,
			pip->val_1 + i,
			pip->val_2 + i);
		if (j == 0) {
			pip->dim = i;
			continue;
		}
		if (j > 0 &&
			(j != 7 || fabs(foo-i*pip->step) > STEP_EPSILON))
		{
			fprintf(stderr, "read_pair_int(): bad format at line %d\n",
				i);
			free_pip(pip);
			return NULL;
		}
		fgets(buf, BUF_SIZE, f);
	}

	/* Doplnime prvni radek tabulky */
	pip->val_m12[0] = 2*pip->val_m12[1];
	pip->val_m10[0] = 2*pip->val_m10[1];
	pip->val_m6[0] = 2*pip->val_m6[1];
	pip->val_m2[0] = 2*pip->val_m2[1];
	pip->val_m1[0] = 2*pip->val_m1[1];
	pip->val_1[0] = 0.0;
	pip->val_2[0] = 0.0;

	fclose(f);
	return pip;
}

static void free_pip(pip)
pair_int *pip;
{
	free(pip->val_m12);
	free(pip->val_m10);
	free(pip->val_m6);
	free(pip->val_m2);
	free(pip->val_m1);
	free(pip->val_1);
	free(pip->val_2);

	free(pip);
}

/* --------------------------------------------------------------------*/
/* Funkce pro praci s tabulkou torznich interakci */

static dih_int *read_dihfile(template, lambda)
char *template;
REAL lambda;
{
	char buf[BUF_SIZE];
	dih_int *dip, diharray[MAX_DIHS];
	REAL foo, l1;
	FILE *f;
	int i;

	if (!template)
		template = DEFAULT_DIHFILE_TEMPLATE;

	sprintf(buf, template, lambda);
	if ((f = fopen(buf, "r")) == NULL) {
		fprintf(stderr, "read_dihfile(): cannot open file %s: %s\n",
			buf, strerror(errno));
		return NULL;
	}

	/* Hlavicka souboru */
	if (fscanf(f, "lambda=%lf, acc=%lf, eps=%lf, maxpts=%lf\n", &l1,
		&foo, &foo, &foo) != 4 || fabs(lambda - l1) > LAMBDA_EPSILON) { 
		fprintf(stderr, "read_dihfile(): inconsistent header\n");
		fprintf(stderr, "lambda = %f, l1=%f\n", lambda, l1);
		return NULL;
	}

	/* Cteme postupne tabulky pro jednotlive torze */
	for(i=0; read_dip(diharray+i, f, lambda); i++)
		;
	if (!i)
		return NULL;

	fclose(f);
	/* Potrebujeme jeste jeden prvek jako ukoncovadlo. Konec pole
	 * oznacime tak, ze dip[a].title bude NULL pro posledni a.
	 */
	i++;
#ifdef DEBUG
	fprintf(stderr, "%d\n",i);
#endif
	dip = malloc(sizeof(dih_int)*i);
	memcpy(dip, diharray, sizeof(dih_int)*i);
	dip[i-1].title = NULL;
	return dip;
}

static int read_dip(dip, f, lambda)
dih_int *dip;
FILE *f;
REAL lambda;
{
	char buf[BUF_SIZE];
	int i, x, y, z;
	REAL t, a1, a2;

	if (fscanf(f, "%11c %d %*f %d %d %d %*f %*f %*f %*s \n",
		buf, &(dip->n), &(dip->tdim), &(dip->adim1), &(dip->adim2)) != 5) {
		if (!feof(f)) 
			fprintf(stderr, "read_dip(): Incorrect header\n");
		return 0;
	}

	buf[11] = '\0';
#ifdef DEBUG
	fprintf(stderr ,"<%s>\n", buf);
#endif
	dip->title = strdup(buf);
	dip->val = malloc(sizeof(REAL)*dip->tdim*dip->adim1*dip->adim2);

	for (i=x=0; x<dip->tdim; x++)
	for (y=0; y<dip->adim1; y++)
	for (z=0; z<dip->adim2; z++,i++)
	{
		if (fscanf(f, "%lf %lf %lf %lf", &t, &a1, &a2,
			dip->val+i) != 4) {
			fprintf (stderr, "read_dip(): format error at line %d\n", i);
			return 0;
		}
		fgets(buf, BUF_SIZE, f);/* skip to \n */
#ifdef DEBUG
		fprintf(stderr, "value is %lf\n", dip->val[i]);
#endif
		if ((fabs(t-x*M_PI/(dip->tdim-1)) > STEP_EPSILON)
			|| (fabs(a1-y*M_PI/(dip->adim1-1)) > STEP_EPSILON)
			|| (fabs(a2-z*M_PI/(dip->adim2-1)) > STEP_EPSILON))
		{
			fprintf (stderr, "read_dip(): invalid coordinates at %d\n",
				i);
			fprintf (stderr, "t is %f, should be %f\n", t, x*M_PI/(dip->tdim-1));
			fprintf (stderr, "a1 is %f, should be %f\n", a1, y*M_PI/(dip->adim1-1));
			fprintf (stderr, "a2 is %f, should be %f\n", a2, z*M_PI/(dip->adim2-1));
			return 0;
		}
	} /* Tady konci tri for-cykly */

	return 1;
}

static void free_dip(dip)
dih_int *dip;
{
	dih_int *dip1;
	for(dip1=dip; dip->title; dip++) {
		free(dip->title);
		free(dip->val);
	}
	free(dip1);
}

static int compute_tdihnum(ffp)
force_field *ffp;
{
	dih_int *dip = ffp->dip;
	int i,j;
	char buf[BUF_SIZE];

	for (i=0; i<ffp->ndihs; i++)
	{
		sprintf(buf,"%.2s-%.2s-%.2s-%.2s",
			ffp->atypenames[ffp->dihat1[i]],
			ffp->atypenames[ffp->dihat2[i]],
			ffp->atypenames[ffp->dihat3[i]],
			ffp->atypenames[ffp->dihat4[i]]);

		for (j=0; dip[j].title; j++)
			if (!strcmp(buf, dip[j].title)
				&& dip[j].n == ffp->pn[ffp->dihnum[i]])
				break;
		if (!dip[j].title) {
			fprintf(stderr, "compute_tdihnum(): Cannot found dihedral %s\n", buf);
			return 0;
		}
		ffp->tdihnum[i]=j;
	}
	return 1;
}

/* --------------------------------------------------------------------*/
/* Funkce pro praci s tabulkou uhlovych interakci */

static angle_int *read_anglefile(template, lambda)
char *template;
REAL lambda;
{
	char buf[BUF_SIZE];
	angle_int *aip, anglearray[MAX_ANGLES];
	REAL foo, l1;
	FILE *f;
	int i;

	if (!template)
		template = DEFAULT_ANGLEFILE_TEMPLATE;
	sprintf(buf, template, lambda);

	if ((f = fopen(buf, "r")) == NULL) {
		fprintf(stderr, "read_anglefile(): cannot open file %s: %s\n",
			buf, strerror(errno));
		return NULL;
	}

	/* Hlavicka souboru */
	if (fscanf(f, "lambda=%lf, acc=%lf, eps=%lf, maxpts=%lf\n", &l1,
		&foo, &foo, &foo) != 4 || fabs(lambda - l1) > LAMBDA_EPSILON) { 
		fprintf(stderr, "read_anglefile(): inconsistent header\n");
		fprintf(stderr, "lambda = %f, l1=%f\n", lambda, l1);
		return NULL;
	}

	for(i=0; read_aip(anglearray+i, f, lambda); i++)
		;
	if (!i)
		return NULL;

	fclose(f);
	i++;
#ifdef DEBUG
	fprintf(stderr, "%d\n",i);
#endif
	aip = malloc(sizeof(angle_int)*i);
	memcpy(aip, anglearray, sizeof(angle_int)*i);
	aip[i-1].title = NULL;
	return aip;
}

static int read_aip(aip, f, lambda)
angle_int *aip;
FILE *f;
REAL lambda;
{
	char buf[BUF_SIZE];
	int i;
	REAL a;

	if (fscanf(f, "%8c %*f %d %*f %*f\n", buf, &(aip->dim)) != 2) {
		if (!feof(f)) 
			fprintf(stderr, "read_aip(): Incorrect header\n");
		return 0;
	}

	buf[8] = '\0';
#ifdef DEBUG
	fprintf(stderr ,"<%s>\n", buf);
#endif
	aip->title = strdup(buf);
	aip->val = malloc(sizeof(REAL)*aip->dim);

	for (i=0; i<aip->dim; i++)
	{
		if (fscanf(f, "%lf %*f %*f %lf", &a, aip->val+i) != 2) {
			fprintf (stderr, "read_aip(): format error at line %d\n", i);
			return 0;
		}
		fgets(buf, BUF_SIZE, f);/* skip to \n */
#ifdef DEBUG
		fprintf(stderr, "value is %lf\n", aip->val[i]);
#endif
		if (fabs(a-i*M_PI/(aip->dim-1)) > STEP_EPSILON)
		{
			fprintf (stderr, "read_aip(): invalid coordinates at %d\n",
				i);
			fprintf (stderr, "a is %f, should be %f\n", a, i*M_PI/(aip->dim-1));
			return 0;
		}
	}

	return 1;
}

static void free_aip(aip)
angle_int *aip;
{
	angle_int *aip1;
	for(aip1=aip; aip->title; aip++) {
		free(aip->title);
		free(aip->val);
	}
	free(aip1);
}

static int compute_tanglenum(ffp)
force_field *ffp;
{
	angle_int *aip = ffp->aip;
	int i,j;
	char buf[BUF_SIZE];

	for (i=0; i<ffp->nangles; i++)
	{
		sprintf(buf,"%.2s-%.2s-%.2s",
			ffp->atypenames[ffp->angleat1[i]],
			ffp->atypenames[ffp->angleat2[i]],
			ffp->atypenames[ffp->angleat3[i]]);

		for (j=0; aip[j].title; j++)
			if (!strcmp(buf, aip[j].title))
				break;
		if (!aip[j].title) {
			fprintf(stderr, "compute_tanglenum: Cannot found anglee %s\n", buf);
			return 0;
		}
		ffp->tanglenum[i]=j;
	}
	return 1;
}

/* Funkce main() pro ladici ucely */
#ifdef DEBUG
int main()
{
	force_field *ffp;
	angle_int *aip;
	ffp = prm2ff("../molekuly/alb/alb.top");
	aip = read_anglefile("tabulky/alb.13_", 0.410, NULL);
	compute_anglenum(ffp, aip);
}
#endif /* DEBUG */
